// CoopOS Demo  (C) 2019 Helmut Weber  [ www.HelmutWeber.de ] Demo-3


// Neu: taskDelay(x) erlaubt zeitliche Steuerung in JEDEM Task
// Plötzlich wird es sehr schnell: 600000 Scheduler calls pro Sekunde (Esp8266)




// Schneller ohne WLAN (ESP8266)
#define NO_WLAN

struct tcb {
  void      (*Func)();      // Name des Tasks
  uint32_t    LastCalled;       // Zeit: letzter Aufruf 
  uint32_t    Delay;        // Verzögerung für Task
  char      State;        // Status: READY, DELAYED
};

#define MAXTASKS 10
int    numTasks;

struct tcb      Tasks[MAXTASKS];

enum state { READY, DELAYED };

struct tcb   *thisTask;

// -------------------------------------------------

uint32_t SchedulerCalls;

void Scheduler() {
uint32_t     m;

    for (int i=0; i<numTasks; i++) {
        SchedulerCalls++;
        digitalWrite(15,HIGH);
        digitalWrite(15,LOW);
        m=micros();
        thisTask = &Tasks[i];
        if (thisTask->State == DELAYED) {
            if ((m-thisTask->LastCalled) >= thisTask->Delay) {
                thisTask->State = READY;
            }
            else  { thisTask->State=DELAYED; }
        }

        
        if (thisTask->State == READY) {
            thisTask->Func();                  // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<< call the Task
            //Serial.println(thisTask->Delay);
            thisTask->LastCalled = m;
        }
    } // for
} // Scheduler
// -------------------------------------------------



#define taskBegin()                                                            \
  static int _mark = 0;                                                        \
  switch (_mark) {                                                             \
  case 0:

#define taskEnd()    _mark=0;\
    return;  \
  }


#define taskSwitch() _mark=__LINE__; return;  case __LINE__:

extern struct tcb *thisTask;

#define taskDelay(val) _mark=__LINE__; thisTask->Delay =val; thisTask->State=DELAYED; return; case __LINE__:


void Task1() {
    taskBegin();
    while(1) {
      Serial.println("Task1 -1");
      taskDelay(100000);
      
      Serial.println("Task1 -2");
      taskDelay(200000);
      
      Serial.println("Task1 -3");
      taskDelay(300000);
      
      Serial.println("Task1 -4");
      taskDelay(400000);             // Zusammen 1 Sekunde

      Serial.print("Task-Switches: ");
      taskSwitch();
      Serial.println(SchedulerCalls); SchedulerCalls=0;
      taskSwitch();
    }
    taskEnd();
}



// 10 µs HIGH alle 100 µs = 10 kHz
void Task2() {
  taskBegin();
  while(1) {
    digitalWrite(5,HIGH);
    taskDelay(10);
    digitalWrite(5,LOW);
    taskDelay(90);
  }
  taskEnd();
}




void setup() {
  pinMode(5,OUTPUT);
  pinMode(15,OUTPUT);
  Serial.begin(500000);
  delay(1000);
  for (int i=0; i<50; i++) Serial.println();
  Serial.println("(C) 2019 Helmut Weber [www.HelmutWeber.de]  CoopOS Demo-3\n");
  // -------------------------------------------------
  // Tasks initialisieren:
  Tasks[0].Func =  Task1;
  Tasks[0].State = DELAYED;
  Tasks[0].Delay = 5000000;       // Anfangsverzögerung 5 Sekunden
  Tasks[0].LastCalled = micros();

  Tasks[1].Func =  Task2;
  Tasks[1].State = READY;
  Tasks[1].Delay = 0;
  Tasks[1].LastCalled = micros();

  numTasks=2;
  
  // -------------------------------------------------


}

#ifdef NO_WLAN
void loop() {
  while(1) {
    Scheduler();
    ESP.wdtFeed();
  }
}

#else
void loop() {
  Scheduler();
}
#endif



// Wer es mal ausprobieren möchte:
// Hier das Programm für Linux:
// bei mir 40000000 (Vierzig Millionen) Taskswitches pro Sekunde !
// sudo ./Demo-3

/*

TaskSwitches/s 40599086
Task1 -1
Task1 -2
Task1 -3
Task1 -4
TaskSwitches/s 40669106
Task1 -1
Task1 -2
Task1 -3
Task1 -4
TaskSwitches/s 40444868
Task1 -1
Task1 -2
Task1 -3
Task1 -4
TaskSwitches/s 40717658








// CoopOS Demo  (C) 2019 Helmut Weber  [ www.HelmutWeber.de ] Demo-3


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>


// Neu: taskDelay(x) erlaubt zeitliche Steuerung in JEDEM Task
// Plötzlich wird es sehr schnell: 600000 Scheduler calls pro Sekunde


uint64_t micros() {
    struct timeval tv;
    gettimeofday(&tv,NULL);
    return tv.tv_sec*(uint64_t)1000000+tv.tv_usec;
}

// Schneller ohne WLAN (ESP8266)
#define NO_WLAN

struct tcb {
  void      (*Func)();      // Name des Tasks
  uint32_t    LastCalled;       // Zeit: letzter Aufruf 
  uint32_t    Delay;        // Verzögerung für Task
  char      State;        // Status: READY, DELAYED
};

#define MAXTASKS 10
int    numTasks;

struct tcb      Tasks[MAXTASKS];

enum state { READY, DELAYED };

struct tcb   *thisTask;

// -------------------------------------------------

uint32_t SchedulerCalls;

void Scheduler() {
uint32_t     m;

    for (int i=0; i<numTasks; i++) {
        SchedulerCalls++;
        //digitalWrite(15,HIGH);
        //digitalWrite(15,LOW);
        m=micros();
        thisTask = &Tasks[i];
        if (thisTask->State == DELAYED) {
            if ((m-thisTask->LastCalled) >= thisTask->Delay) {
                thisTask->State = READY;
            }
            else  { thisTask->State=DELAYED; }
        }

        
        if (thisTask->State == READY) {
            thisTask->Func();                  // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<< call the Task
            //printf(thisTask->Delay);
            thisTask->LastCalled = m;
        }
    } // for
} // Scheduler
// -------------------------------------------------



#define taskBegin()                                                            \
  static int _mark = 0;                                                        \
  switch (_mark) {                                                             \
  case 0:

#define taskEnd()    _mark=0;\
    return;  \
  }


#define taskSwitch() _mark=__LINE__; return;  case __LINE__:

extern struct tcb *thisTask;

#define taskDelay(val) _mark=__LINE__; thisTask->Delay =val; thisTask->State=DELAYED; return; case __LINE__:


void Task1() {
    taskBegin();
    while(1) {
      printf("Task1 -1\n");
      taskDelay(100000);
      
      printf("Task1 -2\n");
      taskDelay(200000);
      
      printf("Task1 -3\n");
      taskDelay(300000);
      
      printf("Task1 -4\n");
      taskDelay(400000);             // Zusammen 1 Sekunde

      printf("TaskSwitches/s %d\n",SchedulerCalls); SchedulerCalls=0;
      taskSwitch();
    }
    taskEnd();
}



// 10 µs HIGH alle 100 µs = 10 kHz
void Task2() {
  taskBegin();
  while(1) {
    //digitalWrite(5,HIGH);
    taskDelay(10);
    //digitalWrite(5,LOW);
    taskDelay(90);
  }
  taskEnd();
}




void setup() {
  //pinMode(5,OUTPUT);
  //pinMode(15,OUTPUT);
  //Serial.begin(500000);
  //delay(1000);
  for (int i=0; i<50; i++) printf("\n");
  printf("(C) 2019 Helmut Weber [www.HelmutWeber.de]  CoopOS Demo-3\n");
  // -------------------------------------------------
  // Tasks initialisieren:
  Tasks[0].Func =  Task1;
  Tasks[0].State = DELAYED;
  Tasks[0].Delay = 1000000;       // Anfangsverzögerung 1 Sekunde
  Tasks[0].LastCalled = micros();

  Tasks[1].Func =  Task2;
  Tasks[1].State = READY;
  Tasks[1].Delay = 0;
  Tasks[1].LastCalled = micros();

  numTasks=2;
  
  // -------------------------------------------------


}

#ifdef NO_WLAN
void loop() {
  while(1) {
    Scheduler();
    //ESP.wdtFeed();
  }
}

#else
void loop() {
  Scheduler();
}
#endif


int main() {
  setup();
  while(1) {
    loop();
  }
}


 */







