// CoopOS Demo  (C) 2019 Helmut Weber  [ www.HelmutWeber.de ] Kurs-6


#ifndef COOPOS_H
#define COOPOS_H

#include <Arduino.h>


// Config
//---------------------------------------------------------------------------------
#define MAXTASKS 10
#define MAXRESOURCE 3
#define INTERNAL_PRIO
#define LA
#define NO_WLAN               // Schneller ohne WLAN
//---------------------------------------------------------------------------------

enum state { READY, DELAYED, BLOCKED, WAITSIG, WAITRES };



int             numTasks=1;   // Task0 is reserved for Idle (to do)


struct tcb {
  uint8_t     ID;             // entspricht von InitTask() zurückgegebenem Task-Handle
  char        *Name;          // Name des Tasks als Text
  void        (*Func)();      // Name des Tasks
  uint8_t     Priority;       // Priorität // reserved
  uint32_t    LastCalled;     // Zeit: letzter Aufruf 
  uint32_t    Delay;          // Verzögerung für Task
  state       State;          // Status: READY, DELAYED
  uint8_t     Signal;         // wait for a signal
  uint8_t     Resource;       // wait for a resource
};



// Defines for TaskSwitching
//---------------------------------------------------------------------------------
#define taskBegin()                                                            \
  static int _mark = 0;                                                        \
  switch (_mark) {                                                             \
  case 0:

#define taskEnd()    _mark=0;\
    return;  \
  }


#define taskSwitch() _mark=__LINE__; return;  case __LINE__:

extern struct tcb *thisTask;

#define taskDelay(val)    _mark=__LINE__; thisTask->Delay =val; thisTask->State=DELAYED; return; case __LINE__:

// kann auf vom ISR benutzt werden:
#define taskStop(id)      _mark=__LINE__; Tasks[id]Delay =0;    Task[id].State=BLOCKED; 

#define taskStopMe()      _mark=__LINE__; thisTask->Delay =0;   thisTask->State=BLOCKED; return; case __LINE__:

#define taskResume(id)    _mark=__LINE__; Tasks[id].Delay =0;   Tasks[id].State=READY; return; case __LINE__:

#define taskWaitSig(sig)  _mark=__LINE__; thisTask->Delay =0;   thisTask->Signal=sig; thisTask->State=WAITSIG; return; case __LINE__:

// kann auf vom ISR benutzt werden:
#define taskSetSig(sig)  \
for (int ix=1; ix<numTasks; ix++) {\
struct tcb *tp;\
  tp=&Tasks[ix];\
  if (tp->Signal==sig) {\
    tp->Signal=0; tp->State=READY;\
  }\
}


#define taskWaitRes(res)  _mark=__LINE__; \
/*Serial.print("Wait RES "); Serial.print(res); Serial.print(".. Task " ); Serial.println(thisTask->ID);*/\
/*  Wenn die resource frei ist, dann sofort weiter */\
if (Resource[res]==0) {\
  Resource[res]=thisTask->ID;\
  /*Serial.print("RES is free, Resource "); Serial.println(res);*/\
}\
/*  Wenn die resource NICHTfrei ist: gehört sie mir schon ? */\
else {\
  if (Resource[res]==thisTask->ID) {\
    /*Serial.print("RES is mine, Task "); Serial.println(thisTask->ID);*/\
  }\
  /*  ist fremd besetzt ! Task muss warten */\
  else {\
    /*Serial.print("RES wait, Task "); Serial.println(thisTask->ID);*/\
    thisTask->Resource=res;\
    thisTask->State=WAITRES; \
    taskSwitch();\
  }\
}


// kann nur vom Eigentümer benutzt werden (nicht ISR)

#define taskFreeRes(res)  \
/*Serial.print(thisTask->ID); Serial.print("  frees ... "); Serial.println(Resource[res]); */\
if (Resource[res] == thisTask->ID) {\
  /*Serial.println("FREE - RES is mine");*/\
  Resource[res] = 0;\
  for (int ix=1; ix<numTasks; ix++) {\
  struct tcb *tp;\
    tp=&Tasks[ix];\
    if (tp->State == WAITRES) {\
      if (tp->Resource==res) {\
        /*Serial.print(thisTask->ID); Serial.print(" enables "); Serial.println(tp->ID);*/\
        tp->Resource=0; tp->State=READY;\
        Resource[res]=tp->ID;\
        break;\
      }\
    }\
  }\
}


//---------------------------------------------------------------------------------



// Globals
//---------------------------------------------------------------------------------

struct tcb      Tasks[MAXTASKS];
struct tcb      *thisTask;
volatile uint8_t         Resource[MAXRESOURCE];
//---------------------------------------------------------------------------------


// -------------------------------------------------

void Scheduler() {
uint32_t     m;
extern uint32_t SchedCount;

    for (int i=1; i<numTasks; i++) {
#ifdef LA
        digitalWrite(15,HIGH);
        digitalWrite(15,LOW);
#endif
        SchedCount++;
        m=micros();
        thisTask = &Tasks[i];
        if (thisTask->State == DELAYED) {
            if ((m-thisTask->LastCalled) >= thisTask->Delay) {
                thisTask->State = READY;
            }
            //else  { thisTask->State=DELAYED; }
        }

        
        if (thisTask->State == READY) {
            thisTask->Func();                  // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<< call the Task
            //Serial.println(thisTask->Delay);
            thisTask->LastCalled = m;
#ifdef INTERNALPRIO
            break;
#endif
        }
    } // for
} // Scheduler
// -------------------------------------------------

int TaskInit(char *name, void (*f)(void), uint8_t priority, state stat, uint32_t del) {
  if (numTasks >= MAXTASKS) {
    Serial.println("ERR: MAXTASKS"); 
    return -1;                                // Fehler
  }

  Tasks[numTasks].ID =          numTasks;
  Tasks[numTasks].Name =        name;
  Tasks[numTasks].Func =        f;
  Tasks[numTasks].Priority =    priority;     // reserved
  Tasks[numTasks].State =       stat;
  Tasks[numTasks].Delay =       del;
  if (Tasks[numTasks].Delay) Tasks[numTasks].State =  DELAYED;

  numTasks++;
  return (numTasks-1);
}

// -------------------------------------------------


#endif






