// CoopOS Demo  (C) 2019 Helmut Weber  [ www.HelmutWeber.de ] Demo-6

// Die Demos 0-6 laufen auf JEDEM Prozessor unter der Arduino-IDE (ab AtTiny45) !
// (Ok, ESP.wdtFeed(); auskommentieren !)

// Es ist damit die erste Wahl, um einen Sketch mit einem neuen Prozessor auszuprobieren.
// Die Arduino-IDE verfolgt das Konzept, die Programme für alle Prozessoren möglichst gleich zu lassen.
// CoopOS passt sich dem an.

// Hier sind alle CoopOS-Befehle vorhanden. Demo-6 ist damit der beste Ausgangspunkt für alle Versuche
// mit dem vorhandenen Prozessor.

// Demo-7 und Demo-8 sind (in dieser Form) spezifisch für ESP8266
// Sie lassen sich aber leicht für andere Prozessoren umschreiben



// Alt:
// taskBegin();
// taskEnd();
// taskSwitch();
// taskDelay(x_µs);
// taskStopMe();
// taskResume(id);

// Neu;
// taskWaitSig(sig);      Warten auf ein Signal. Es können mehrere gleichzeitig warten.
// taskSetSig(sig);       Setzen eines Signals. Es können mehrere "geweckt" werden. 
// taskWaitRes();         Wartet auf eine Resource und wartet oder macht gleich weiter, wenn Resource frei ist.
// taskFreeRes(res);      Nur vom Eigentümer gültig! Befreit besetzte Resource.

/*
Hier ist ein Output von einem Aruino NANO:

SchedCount 44525 ~ µs per call 22.45
Task1 -1
Task1 -2
Task1 -3
Task1 -4
SchedCount 44535 ~ µs per call 22.45
Task1 -1
Task1 -2
Task1 -3
Task1 -4
SchedCount 44535 ~ µs per call 22.45
Task1 -1
Task1 -2
Task1 -3
Task1 -4
Task1 -1
SchedCount 44525 ~ µs per call 22.45


and from Esp8266:

SchedCount 645785 ~ µs per call 1.55
Task1 -1
Task1 -2
Task1 -3
Task1 -4
SchedCount 645780 ~ µs per call 1.55
Task1 -1
Task1 -2
Task1 -3
Task1 -4
Task1 -1
SchedCount 645795 ~ µs per call 1.55
Task1 -2
Task1 -3
Task1 -4
Task1 -1
Task1 -2
SchedCount 645770 ~ µs per call 1.55

//----------------------------
10fach höhere Taktfrequenz (160 zu 16 MHz) und 32 statt 8 Bit bei dem Scheduler-Count zum Faktor 14.5,
die ESp8266 schneller ist, als der Arduin NANO.
//----------------------------

*/


#include "CoopOS.h"

// Globals
// TaskHandles:
uint32_t  T_ID1, T_ID2, T_ID3, T_ID4;
uint32_t SchedCount;

// Resources
#define DISPLAY       2           // für Serial Output < MAXRESOURE !!!!
#define TASK_3_4_PIN  1           // für LA Darstellung Task3, Task4

// Pins
#define Task2_Pin     5           // Output Pin genutzt von Task2

// Signals
#define SIG_SYNC      23          // jeder Wert von 1..255

void Task1() {
    taskBegin();
    while(1) {
      taskWaitRes(DISPLAY);
      taskWaitSig(SIG_SYNC); 
      Serial.println("Task1 -1"); taskFreeRes(DISPLAY);         
      taskDelay(100000);
      taskWaitRes(DISPLAY);
      taskWaitSig(SIG_SYNC); 
      Serial.println("Task1 -2"); taskFreeRes(DISPLAY);         
      taskDelay(200000);
      taskWaitRes(DISPLAY);
      taskWaitSig(SIG_SYNC); 
      Serial.println("Task1 -3"); taskFreeRes(DISPLAY);         
      taskDelay(300000);
      taskWaitRes(DISPLAY);
      taskWaitSig(SIG_SYNC); 
      Serial.println("Task1 -4"); taskFreeRes(DISPLAY);         
      taskDelay(300000);
    }
    taskEnd();
}



void Task2() {
  taskBegin();
  while(1) {
    digitalWrite(Task2_Pin, HIGH);                 taskDelay(10);    // 10 µs HIGH
    digitalWrite(Task2_Pin, LOW);                  taskDelay(100);   // 100µs LOW
  }
  taskEnd();
}


void Task3() {
  taskBegin();
  taskStopMe();                               // wait for Task4 to sync
  while(1) {
                    
    //taskStopMe();
    //taskWaitSig(SIG_SYNC);
    //Serial.println("Task3 wait");
    taskWaitRes(TASK_3_4_PIN);                // reserve resoure 1 for ME, so Task4 has to wait for resource
    digitalWrite(12,LOW); 
    taskSetSig(SIG_SYNC);
    //Serial.println("Task3");
    taskDelay(50-2);                          // resource is blocked by me
    //Serial.println("Task3 ready");
    taskFreeRes(TASK_3_4_PIN);                // resource set free after 50 ms
    taskDelay(20);                            // allow Task4 to get resource
  }
  taskEnd();
}


void Task4() {
  taskBegin();
  taskWaitRes(TASK_3_4_PIN);                  // reserve resoure 1 for ME, so Task3 has to wait for resource
  taskResume(T_ID3);                          // reagiert "normalerweise" nach 2,5µs, wenn bei Init Platz 1
  taskDelay(10);
  while(1) {
    //Serial.println("Task4 wait");
    taskWaitRes(TASK_3_4_PIN);                // resource is blocked by me 
    digitalWrite(12,HIGH);
    taskSetSig(SIG_SYNC);
    //Serial.println("Task4");
    //taskResume(T_ID3);                      // reagiert "normalerweise" nach 2,5µs, wenn bei Init Platz 1
    //taskSetSig(SIG_SYNC);                   // reagiert "normalerweise" nach 2,8µs, wenn bei Init Platz 1
    taskDelay(50-5);
    //Serial.println("Task4 ready");
    taskFreeRes(TASK_3_4_PIN);                // reagiert "normalerweise" nach 8,6µs, wenn bei Init Platz 1
    taskDelay(20);                            // allow Task3 to get resource
    
  }
  taskEnd();
}


void Task5() {
static double x;
  taskBegin();
  while(1) {
    // About 670000 TaskSwitches/s :
    taskWaitRes(DISPLAY);
    taskWaitSig(SIG_SYNC);
    Serial.print("SchedCount "); 
    taskWaitSig(SIG_SYNC);
    Serial.print( SchedCount); 
    taskWaitSig(SIG_SYNC);
    Serial.print(" ~ µs per call "); 
    taskWaitSig(SIG_SYNC);
    x=1000000.0/(double)SchedCount;
    taskWaitSig(SIG_SYNC);
    Serial.println(x);
    
    taskFreeRes(DISPLAY);
    
    SchedCount=0;
    taskDelay(1000000);
  }
  taskEnd();
}


// -------------------------------------------------



void setup() {

  pinMode(Task2_Pin, OUTPUT);             // Task2
  pinMode(12,OUTPUT);                     // Task3, Task4
  
#ifdef LA
  pinMode(15,OUTPUT);                     // Logic-Analyzer: Scheduler
#endif
  
  Serial.begin(500000);
  delay(1000);
  for (int i=0; i<50; i++) Serial.println();
  
  Serial.println(F("(C) 2019 Helmut Weber [www.HelmutWeber.de]  CoopOS Demo-6"));
  Serial.println("TaskInit ...");
  
  // -------------------------------------------------
  // Tasks initialisieren:
  //int TaskInit(char *name, void (*f)(void), uint8_t priority, state stat, uint32_t del) 
  T_ID3=TaskInit(       "T3",      Task3,           50,               READY,      0);
  T_ID4=TaskInit(       "T4",      Task4,           50,               READY,      0);
  T_ID1=TaskInit(       "T1",      Task1,           50,               DELAYED,    5000000);
  T_ID2=TaskInit(       "T2",      Task2,           50,               READY,      0);
        TaskInit(       "T5",      Task5,           50,               READY,      0);
  Serial.println("TaskInit ready!");
  
}





#ifdef NO_WLAN
void loop() {
  while(1) {
    Scheduler();
    ESP.wdtFeed(); // nur wichtig für Esp8266
  }
}

#else
void loop() {
  Scheduler();
}
#endif
