// CoopOS Demo  (C) 2019 Helmut Weber  [ www.HelmutWeber.de ] Demo-7 FÜR ARDUINO NANO

// Neu;
// Timer2 Interrupt: Task vom Interrupt starten
// Ein Task, der vom Interrupt geweckt wird, reagiert nach 20-100 µs Synchronisation
// Externer Interrupt an D2
/*

Serial Output (oder auch Ausgabe auf andere Devices) ist bei embedded systems immer ein 
Flaschenhals.

Da taskSetSig() sehr schnell ist, kann mit anderen Tasks synchronisiert werden, um diese
nicht zu stören.
SYNC_TASK_2 sort dafür, dass serielle Ausgaben in dem 100µs LOW Intervall ausgegeben werden.
Damit wird Tasks 2 deterministisch !
    ODER
SYNC_TASK_3_4 sorgt für Synchronisation mit Task 3,4

Natürlich sind von dem Arduino Nano keine Ergebnisse wie vom ESO8266 zu erwarten !
Er ist etwa um den Faktor 14.5 langsamer.

Aber trotzdem kommt nach 20 bis 100 µs - nach warten auf die freie Resource DISPLAY UND die Synchronisation
mit Task_2 der Task_5 zur Ausführung. 

Der Interrupt an D2 (gegen Masse) kommt nach ca. 40 - 150 µs zur Ausführung (mit DISPLAY, SYNC)

Output:
=======

Vom Interrupt: ISR_Count 20000 Delay 10000 Interrupts*100µs, genauer DeltaT 32 µs
Task1 -2
Task1 -3
Task1 -4
Task1 -1
SchedCount 49677 ~ µs per call 20.12
>>> EXT IRQ: (µs)40
Task1 -2
Vom Interrupt: ISR_Count 20000 Delay 10000 Interrupts*100µs, genauer DeltaT 32 µs
Task1 -3
Task1 -4
Task1 -1
>>> EXT IRQ: (µs)36
Task1 -2
SchedCount 49590 ~ µs per call 20.15
Vom Interrupt: ISR_Count 20001 Delay 10001 Interrupts*100µs, genauer DeltaT 68 µs
Task1 -3
>>> EXT IRQ: (µs)72
Task1 -4
>>> EXT IRQ: (µs)36
Task1 -1
>>> EXT IRQ: (µs)80
Task1 -2
SchedCount 49458 ~ µs per call 20.21
Vom Interrupt: ISR_Count 20000 Delay 10000 Interrupts*100µs, genauer DeltaT 16 µs





 */

#define   SYNC_TASK_2           // weniger Jitter bei Task2
//#define SYNC_TASK_3_4         // weniger jitter bei Task3,4




#include "CoopOS.h"
#include "pins.h"


// Globals
Sched Sched;
uint8_t inIrq;
static uint32_t irq_time;

// TaskHandles:
uint32_t  T_ID1, T_ID2, T_ID3, T_ID4, T_ID5, T_ID6, T_ID7;
uint32_t SchedCount;

// Resources
#define DISPLAY       2           // für Serial Output < MAXRESOURE !!!!

// Signals
#define SIG_SYNC      23              // jeder Wert von 1..255

#define SIG_FROM_IRG      99          // alle 10000 Interrupts

#define EXT_IRQ           63          



/* Test Pin */
#define PIN_HIGH 1


extern "C" {
uint32_t ISR_Count, CCountStart, ThisTick, LastTick, DeltaTick;
}

/*********************************************************
 * Function Name   : Timer_Interrupt()
 * Description     : Toggles the GPIO_ISR_PIN every x usec
 * Returns         : NONE
 * Params          : NONE
 ********************************************************/
 
//void Timer_Interrupt()
ISR(TIMER2_COMPA_vect){
static unsigned char pin_status;
    ISR_Count++;
    if (ISR_Count == 20000) {
      taskSetSig(SIG_FROM_IRG);
      CCountStart=micros();
    }
    //pin_status = digitalRead(GPIO_ISR_PIN);
    pin_status++;
    pin_status &= 1;
    if (pin_status == PIN_HIGH)
        digitalWrite(GPIO_ISR_PIN, LOW);
        //GPOC = (1 << GPIO_ISR_PIN);
    else
        digitalWrite(GPIO_ISR_PIN, HIGH);
        //GPOS = (1 << GPIO_ISR_PIN);
        
}

// --------------------------------------------------------------------------------------
// End Timer Interrupt


// Externer Interrupt an Pin D2

void ExtIrq() { 
  if (inIrq) return;
  
  if (inIrq==0) {
    inIrq=1;
    irq_time=micros();
    //GPOC = (1 << TASK_3_4_PIN);
    //GPOS = (1 << TASK_3_4_PIN);
    taskSetSig(EXT_IRQ);
    //GPOC = (1 << TASK_3_4_PIN);
  }
  //GPOC = (1 << 13);
}


// Timer2 initialisieren: alle 50 µs

void init_timer2() {
//set timer2 interrupt at 20kHz == 50 µs
cli();
  TCCR2A = 0;// set entire TCCR2A register to 0
  TCCR2B = 0;// same for TCCR2B
  TCNT2  = 0;//initialize counter value to 0
  // set compare match register for 20khz increments
  OCR2A = 99;// = (16*10^6) / (10000*8) - 1 (must be <256)
  // turn on CTC mode
  TCCR2A |= (1 << WGM21);
  // Set CS21 bit for 8 prescaler
  TCCR2B |= (1 << CS21);   
  // enable timer compare interrupt
  TIMSK2 |= (1 << OCIE2A);
sei();
}



void Task1() {
    taskBegin();
    while(1) {
      taskWaitRes(DISPLAY);
      taskWaitSig(SIG_SYNC); 
      Serial.println("Task1 -1"); taskFreeRes(DISPLAY);         
      taskDelay(100000);
      taskWaitRes(DISPLAY);
      taskWaitSig(SIG_SYNC); 
      Serial.println("Task1 -2"); taskFreeRes(DISPLAY);         
      taskDelay(200000);
      taskWaitRes(DISPLAY);
      taskWaitSig(SIG_SYNC); 
      Serial.println("Task1 -3"); taskFreeRes(DISPLAY);         
      taskDelay(300000);
      taskWaitRes(DISPLAY);
      taskWaitSig(SIG_SYNC); 
      Serial.println("Task1 -4"); taskFreeRes(DISPLAY);         
      taskDelay(300000);
    }
    taskEnd();
}




void Task2() {
  taskBegin();
  while(1) {
    digitalWrite(Task2_Pin, HIGH);                 
    taskDelay(50-14);                            // 50 µs HIGH
    digitalWrite(Task2_Pin, LOW);   

#ifdef SYNC_TASK_2             
    taskSetSig(SIG_SYNC);                     // Alternativ kann auf Task3,4 synchronisiert werden
#endif
  
    taskDelay(500);                           // 500µs LOW
  }
  taskEnd();
}


void Task3() {
  taskBegin();
  taskStopMe();                               // wait for Task4 to sync
  while(1) {
                    
    //taskStopMe();
    //taskWaitSig(SIG_SYNC);
    //Serial.println("Task3 wait");
    taskWaitRes(TASK_3_4_PIN);                // reserve resoure 1 for ME, so Task4 has to wait for resource
    digitalWrite(TASK_3_4_PIN,LOW);

#ifdef SYNC_TASK_3_4             
    taskSetSig(SIG_SYNC);                   // alternative Synchronisation                   
#endif    
    //Serial.println("Task3");
    taskDelay(200-2);                          // resource is blocked by me
    //Serial.println("Task3 ready");
    taskFreeRes(TASK_3_4_PIN);                // resource set free after 100 ms
    taskDelay(50);                            // allow Task4 to get resource
  }
  taskEnd();
}


void Task4() {
  taskBegin();
  taskWaitRes(TASK_3_4_PIN);                  // reserve resoure 1 for ME, so Task3 has to wait for resource
  taskResume(T_ID3);                          // reagiert "normalerweise" nach 2,5µs, wenn bei Init Platz 1
  taskDelay(10);
  while(1) {
    //Serial.println("Task4 wait");
    taskWaitRes(TASK_3_4_PIN);                // resource is blocked by me 
    digitalWrite(TASK_3_4_PIN,HIGH);

#ifdef SYNC_TASK_3_4             
    taskSetSig(SIG_SYNC);                     // alternative Synchronisation
#endif

    taskDelay(100);
    //Serial.println("Task4 ready");
    taskFreeRes(TASK_3_4_PIN);                
    taskDelay(50);                            
    
  }
  taskEnd();
}


void Task5() {
static double x;
  taskBegin();
  while(1) {
    // About 670000 TaskSwitches/s :
    taskWaitRes(DISPLAY);
    taskWaitSig(SIG_SYNC);
    Serial.print("SchedCount "); 
    taskWaitSig(SIG_SYNC);
    Serial.print( SchedCount); 
    taskWaitSig(SIG_SYNC);
    Serial.print(" ~ µs per call "); 
    taskWaitSig(SIG_SYNC);
    x=1000000.0/(double)SchedCount;
    taskWaitSig(SIG_SYNC);
    Serial.print(x);
    
    
    taskWaitSig(SIG_SYNC);
    Serial.println();
    
    taskFreeRes(DISPLAY);
    
    SchedCount=0;
    taskDelay(1000000-370);
  }
  taskEnd();
}


void Task6() {
static uint32_t x;
static uint32_t CCountEnd, icount;
  taskBegin();
  while(1) {
    // About 670000 TaskSwitches/s :
    
    
    taskWaitSig(SIG_FROM_IRG);
    CCountEnd=micros();
    icount=ISR_Count;
    cli();
    ISR_Count=0;
    sei();
    
    taskWaitRes(DISPLAY);
    taskWaitSig(SIG_SYNC);
    Serial.print("Vom Interrupt: ISR_Count "); 
    taskDelay(300);
    taskWaitSig(SIG_SYNC);
    Serial.print(icount); 
    taskDelay(300);
    taskWaitSig(SIG_SYNC);
    Serial.print(" Delay "); 
    taskDelay(300);
    taskWaitSig(SIG_SYNC);
    Serial.print(icount-10000); 
    taskDelay(300);
    taskWaitSig(SIG_SYNC);
    Serial.print(" Interrupts*100µs,"); 
    taskDelay(300);
    taskWaitSig(SIG_SYNC);
    Serial.print(" genauer DeltaT "); 
    taskDelay(300);
    taskWaitSig(SIG_SYNC);
    x=(CCountEnd-CCountStart);
    taskWaitSig(SIG_SYNC);
    Serial.print(x); 
    taskWaitSig(SIG_SYNC);
    Serial.println(" µs");

    // Um 1/160µs Ticks anzuzeigen:
//    taskDelay(100);
//    taskWaitSig(SIG_SYNC);
//    Serial.print("Interrupt DeltaT = 5µs in 1/160 µs"); 
//    taskDelay(100);
//    taskWaitSig(SIG_SYNC);
//    Serial.print(" = 6.25 Nano-Sekunden Ticks: "); 
//    taskDelay(100);
//    taskWaitSig(SIG_SYNC);
//    Serial.println(DeltaTick); 
    
    taskFreeRes(DISPLAY);
    
 
    //taskDelay(1000000-370);
  }
  taskEnd();
}




void Task7() {
static uint32_t irq_time2;
  taskBegin();
  while(1) {
    taskWaitSig(EXT_IRQ); 
    irq_time2=micros();        
    inIrq=1;     
    taskWaitRes(DISPLAY);
    Serial.print(">>> EXT IRQ: (µs)");
    Serial.println(irq_time2-irq_time);
    taskDelay(50000);                   // debounce 50 ms
    
    inIrq=0;
    taskFreeRes(DISPLAY);
  }
  taskEnd();
}


// -------------------------------------------------




void setup() {

  pinMode(Task2_Pin, OUTPUT);                      // Task2
  pinMode(TASK_3_4_PIN,OUTPUT);                    // Task3, Task4
  pinMode(GPIO_ISR_PIN,OUTPUT);                    // Timer IRQ
  pinMode(EXT_IRQ_PIN, INPUT_PULLUP);              // Externer Interrupt 
  
#ifdef LA
  pinMode(LA_PIN,OUTPUT);                          // Logic-Analyzer: Scheduler
#endif
  
  Serial.begin(500000);
  for (int i=0; i<50; i++) Serial.println();
  
  Serial.println("(C) 2019 Helmut Weber [www.HelmutWeber.de]  CoopOS Demo-7");
  Serial.println("TaskInit ...");
  
  // -------------------------------------------------
  // Tasks initialisieren:
  //int TaskInit(char *name, void (*f)(void), uint8_t priority, state stat, uint32_t del) 

  T_ID7=Sched.TaskInit(       "T7",      Task7,           50,               READY,      0);
  
#ifdef SYNC_TASK_2
        Sched.TaskInit(       "T6",      Task6,           50,               READY,      0);
  T_ID2=Sched.TaskInit(       "T2",      Task2,           50,               READY,      0);
  T_ID3=Sched.TaskInit(       "T3",      Task3,           50,               READY,      0);
  T_ID4=Sched.TaskInit(       "T4",      Task4,           50,               READY,      0);
        
  T_ID1=Sched.TaskInit(       "T1",      Task1,           50,               DELAYED,    5000000);
#endif



#ifdef SYNC_TASK_3_4
        Sched.TaskInit(       "T6",      Task6,           50,               READY,      0);
  T_ID3=Sched.TaskInit(       "T3",      Task3,           50,               READY,      0);
  T_ID4=Sched.TaskInit(       "T4",      Task4,           50,               READY,      0);
  T_ID2=Sched.TaskInit(       "T2",      Task2,           50,               READY,      0);
  T_ID1=Sched.TaskInit(       "T1",      Task1,           50,               DELAYED,    5000000);
#endif
        Sched.TaskInit(       "T5",      Task5,           50,               READY,      0);
        
  Serial.println("TaskInit ready!");
  Serial.println("Init Timer1 Interrupt 10 µs");
  init_timer2();

  Serial.println("External Interrupt Pin D2");
  pinMode(EXT_IRQ_PIN, INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(EXT_IRQ_PIN), ExtIrq, FALLING);
  

}





#ifdef NO_WLAN
void loop() {
  while(1) {
    Sched.Scheduler();
    
  }
}

#else
void loop() {
  Scheduler();
}
#endif
