// CoopOS Demo  (C) 2019 Helmut Weber  [ www.HelmutWeber.de ] Demo-7

// Neu;
// Timer1 Interrupt: Task vom Interrupt starten
// Ein Task, der vom Interrupt geweckt wird, reagiert nach 2-4 µs
// Synchronisation

/*

Serial Output (oder auch Ausgabe auf andere Devices) ist bei embedded systems immer ein 
Flaschenhals.
Hier dauert dir längste Ausgabe 68 µs.
Da taskSetSig() sehr schnell ist, kann mit anderen Tasks synchronisiert werden, um diese
nicht zu stören.
SYNC_TASK_2 sort dafür, dass serielle Ausgaben in dem 100µs LOW Intervall ausgegeben werden.
Damit wird Tasks 2 deterministisch !
    ODER
SYNC_TASK_3_4 sorgt für Synchronisation mit Task 3,4

 */

#define   SYNC_TASK_2           // weniger Jitter bei Task2
//#define SYNC_TASK_3_4         // weniger jitter bei Task3,4




#include "CoopOS.h"
#include "pins.h"


// Globals
Sched Sched;

// TaskHandles:
uint32_t  T_ID1, T_ID2, T_ID3, T_ID4, T_ID5, T_ID6, T_ID7;
uint32_t SchedCount;

// Resources
#define DISPLAY       2           // für Serial Output < MAXRESOURE !!!!

// Signals
#define SIG_SYNC      23          // jeder Wert von 1..255

#define SIG_FROM_IRG      99          // alle 200000 Interrupts

#define EXT_IRQ           63          // alle 200000 Interrupts


// read cpu ticks
static inline uint32_t asm_ccount(void) {
    uint32_t r;
    asm volatile ("rsr %0, ccount" : "=r"(r));
    return r;
}


// --------------------------------------------------------------------------------------
// Timer Interrupt (NMI, highest priority !)
// --------------------------------------------------------------------------------------
// in timer.c
extern "C" {
void ICACHE_FLASH_ATTR Initialize_ISR(uint32_t mics);
}

/* Test Pin */
#define PIN_HIGH 1

/*********************************************************
 * Function Name   : Timer_NMI_Interrupt()
 * Description     : Toggles the GPIO_ISR_PIN every x usec
 * Returns         : NONE
 * Params          : NONE
 ********************************************************/
extern "C" {
uint32_t ISR_Count, CCountStart, ThisTick, LastTick, DeltaTick;
}

extern "C" void ICACHE_RAM_ATTR Timer_Interrupt()
{
static unsigned char pin_status;
    ISR_Count++;

    // Um 1/160 Ticks anzuzeigen
    //ThisTick=asm_ccount();
    //DeltaTick=ThisTick-LastTick;
    //LastTick=ThisTick;
    
    
    if (ISR_Count == 100000) {
      taskSetSig(SIG_FROM_IRG);
      CCountStart=asm_ccount();
    }
    //pin_status = digitalRead(GPIO_ISR_PIN);
    pin_status++;
    pin_status &= 1;
    if (pin_status == PIN_HIGH)
        //digitalWrite(GPIO_ISR_PIN, LOW);
        GPOC = (1 << GPIO_ISR_PIN);
    else
        //digitalWrite(GPIO_ISR_PIN, HIGH);
        GPOS = (1 << GPIO_ISR_PIN);
        
}

// --------------------------------------------------------------------------------------
// End Timer Interrupt



uint8_t inIrq=0;

uint32_t irq_time;

void ICACHE_RAM_ATTR ExtIrq() {
  if (inIrq) return;
  
  
  if (inIrq==0) {
    inIrq=1;
    irq_time=asm_ccount();
    GPOC = (1 << TASK_3_4_PIN);
    GPOS = (1 << TASK_3_4_PIN);
    taskSetSig(EXT_IRQ);
    GPOC = (1 << TASK_3_4_PIN);
    //detachInterrupt(digitalPinToInterrupt(EXT_IRQ_PIN));
  }
  //GPOC = (1 << 13);
}





void Task1() {
    taskBegin();
    while(1) {
      taskWaitRes(DISPLAY);
      taskWaitSig(SIG_SYNC); 
      Serial.println("Task1 -1"); taskFreeRes(DISPLAY);         
      taskDelay(100000);
      taskWaitRes(DISPLAY);
      taskWaitSig(SIG_SYNC); 
      Serial.println("Task1 -2"); taskFreeRes(DISPLAY);         
      taskDelay(200000);
      taskWaitRes(DISPLAY);
      taskWaitSig(SIG_SYNC); 
      Serial.println("Task1 -3"); taskFreeRes(DISPLAY);         
      taskDelay(300000);
      taskWaitRes(DISPLAY);
      taskWaitSig(SIG_SYNC); 
      Serial.println("Task1 -4"); taskFreeRes(DISPLAY);         
      taskDelay(300000);
    }
    taskEnd();
}



void Task2() {
  taskBegin();
  while(1) {
    digitalWrite(Task2_Pin, HIGH);                 
    taskDelay(20);                            // 20 µs HIGH
    digitalWrite(Task2_Pin, LOW);   

#ifdef SYNC_TASK_2             
    taskSetSig(SIG_SYNC);                     // Alternativ kann auf Task3,4 synchronisiert werden
#endif
  
    taskDelay(100);   // 100µs LOW
  }
  taskEnd();
}


void Task3() {
  taskBegin();
  taskStopMe();                               // wait for Task4 to sync
  while(1) {
                    
    //taskStopMe();
    //taskWaitSig(SIG_SYNC);
    //Serial.println("Task3 wait");
    taskWaitRes(TASK_3_4_PIN);                // reserve resoure 1 for ME, so Task4 has to wait for resource
    digitalWrite(12,LOW);

#ifdef SYNC_TASK_3_4             
    taskSetSig(SIG_SYNC);                   // alternative Synchronisation                   
#endif    
    //Serial.println("Task3");
    taskDelay(70-2);                          // resource is blocked by me
    //Serial.println("Task3 ready");
    taskFreeRes(TASK_3_4_PIN);                // resource set free after 100 ms
    taskDelay(20);                            // allow Task4 to get resource
  }
  taskEnd();
}


void Task4() {
  taskBegin();
  taskWaitRes(TASK_3_4_PIN);                  // reserve resoure 1 for ME, so Task3 has to wait for resource
  taskResume(T_ID3);                          // reagiert "normalerweise" nach 2,5µs, wenn bei Init Platz 1
  taskDelay(10);
  while(1) {
    //Serial.println("Task4 wait");
    taskWaitRes(TASK_3_4_PIN);                // resource is blocked by me 
    digitalWrite(12,HIGH);

#ifdef SYNC_TASK_3_4             
    taskSetSig(SIG_SYNC);                     // alternative Synchronisation
#endif

    taskDelay(70-5);
    //Serial.println("Task4 ready");
    taskFreeRes(TASK_3_4_PIN);                
    taskDelay(20);                            
    
  }
  taskEnd();
}


void Task5() {
static double x;
  taskBegin();
  while(1) {
    // About 670000 TaskSwitches/s :
    taskWaitRes(DISPLAY);
    taskWaitSig(SIG_SYNC);
    Serial.print("SchedCount "); 
    taskWaitSig(SIG_SYNC);
    Serial.print( SchedCount); 
    taskWaitSig(SIG_SYNC);
    Serial.print(" ~ µs per call "); 
    taskWaitSig(SIG_SYNC);
    x=1000000.0/(double)SchedCount;
    taskWaitSig(SIG_SYNC);
    Serial.print(x);
    
    
    taskWaitSig(SIG_SYNC);
    Serial.println();
    
    taskFreeRes(DISPLAY);
    
    SchedCount=0;
    taskDelay(1000000-370);
  }
  taskEnd();
}


void Task6() {
static float x;
static uint32_t CCountEnd, icount;
  taskBegin();
  while(1) {
    // About 670000 TaskSwitches/s :
    
    
    taskWaitSig(SIG_FROM_IRG);
    CCountEnd=asm_ccount();  // 160
    icount=ISR_Count;
    cli();
    ISR_Count=0;
    sei();
    
    taskWaitRes(DISPLAY);
    taskWaitSig(SIG_SYNC);
    Serial.print("Vom Interrupt: ISR_Count "); 
    taskDelay(300);
    taskWaitSig(SIG_SYNC);
    Serial.print(icount); 
    taskDelay(300);
    taskWaitSig(SIG_SYNC);
    Serial.print(" Delay "); 
    taskDelay(300);
    taskWaitSig(SIG_SYNC);
    Serial.print(icount-100000); 
    taskDelay(300);
    taskWaitSig(SIG_SYNC);
    Serial.print(" Interrupts*5µs,"); 
    taskDelay(300);
    taskWaitSig(SIG_SYNC);
    Serial.print(" genauer DeltaT "); 
    taskDelay(300);
    taskWaitSig(SIG_SYNC);
    x=(float)(CCountEnd-CCountStart);
    taskSwitch();
    x/=160.0;
    taskSwitch();
    taskWaitSig(SIG_SYNC);
    Serial.print(x); 
    taskWaitSig(SIG_SYNC);
    Serial.println(" µs");

    // Um 1/160µs Ticks anzuzeigen:
//    taskDelay(100);
//    taskWaitSig(SIG_SYNC);
//    Serial.print("Interrupt DeltaT = 5µs in 1/160 µs"); 
//    taskDelay(100);
//    taskWaitSig(SIG_SYNC);
//    Serial.print(" = 6.25 Nano-Sekunden Ticks: "); 
//    taskDelay(100);
//    taskWaitSig(SIG_SYNC);
//    Serial.println(DeltaTick); 
    
    taskFreeRes(DISPLAY);
    
 
    //taskDelay(1000000-370);
  }
  taskEnd();
}




void Task7() {
static uint32_t irq_time2;
  taskBegin();
  while(1) {
    taskWaitSig(EXT_IRQ); 
    irq_time2=asm_ccount();        
    inIrq=1;     
    taskWaitRes(DISPLAY);
    Serial.print(">>> EXT IRQ: (µs)");
    Serial.println((float)(irq_time2-irq_time)*0.0065 );
    taskDelay(50000);                   // debounce 50 ms
    //attachInterrupt(digitalPinToInterrupt(EXT_IRQ_PIN), ExtIrq, FALLING);
    inIrq=0;
    taskFreeRes(DISPLAY);
  }
  taskEnd();
}


// -------------------------------------------------




void setup() {

  pinMode(Task2_Pin, OUTPUT);                      // Task2
  pinMode(12,OUTPUT);                              // Task3, Task4
  
#ifdef LA
  pinMode(15,OUTPUT);                     // Logic-Analyzer: Scheduler
#endif

  pinMode(EXT_IRQ_PIN, INPUT_PULLUP);
  
  Serial.begin(500000);
  for (int i=0; i<50; i++) Serial.println();
  
  Serial.println("(C) 2019 Helmut Weber [www.HelmutWeber.de]  CoopOS Demo-7");
  Serial.println("TaskInit ...");
  
  // -------------------------------------------------
  // Tasks initialisieren:
  //int TaskInit(char *name, void (*f)(void), uint8_t priority, state stat, uint32_t del) 

  T_ID7=Sched.TaskInit(       "T7",      Task7,           50,               READY,      0);
  
#ifdef SYNC_TASK_2
        Sched.TaskInit(       "T6",      Task6,           50,               READY,      0);
  T_ID2=Sched.TaskInit(       "T2",      Task2,           50,               READY,      0);
  T_ID3=Sched.TaskInit(       "T3",      Task3,           50,               READY,      0);
  T_ID4=Sched.TaskInit(       "T4",      Task4,           50,               READY,      0);
        
  T_ID1=Sched.TaskInit(       "T1",      Task1,           50,               DELAYED,    5000000);
#endif



#ifdef SYNC_TASK_3_4
        Sched.TaskInit(       "T6",      Task6,           50,               READY,      0);
  T_ID3=Sched.TaskInit(       "T3",      Task3,           50,               READY,      0);
  T_ID4=Sched.TaskInit(       "T4",      Task4,           50,               READY,      0);
  T_ID2=Sched.TaskInit(       "T2",      Task2,           50,               READY,      0);
  T_ID1=Sched.TaskInit(       "T1",      Task1,           50,               DELAYED,    5000000);
#endif
        Sched.TaskInit(       "T5",      Task5,           50,               READY,      0);
        
  Serial.println("TaskInit ready!");
  Serial.println("Init Timer1 Interrupt 10 µs");
  Initialize_ISR(10);             // 5 µs sind möglich !

  Serial.println("External Interrupt Pin x");
  pinMode(EXT_IRQ_PIN, INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(EXT_IRQ_PIN), ExtIrq, FALLING);
}





#ifdef NO_WLAN
void loop() {
  while(1) {
    Sched.Scheduler();
    ESP.wdtFeed();
  }
}

#else
void loop() {
  Scheduler();
}
#endif
